//
//  BlockBlockKext.c
//  BlockBlockKext
//
//  Created by Patrick Wardle on 11/11/15.
//  Copyright © 2015 Objective-See. All rights reserved.
//

#include "consts.h"

#include <sys/proc.h>
#include <sys/vnode.h>
#include <security/mac.h>
#include <sys/kern_event.h>
#include <libkern/libkern.h>
#include <mach/mach_types.h>
#include <security/mac_policy.h>
#include <security/mac_framework.h>

/* manual load/unload instructions */

//sudo chown -R root:wheel ~/Desktop/BlockBlockKext.kext
//sudo kextload ~/Desktop/BlockBlockKext.kext
//sudo kextunload -b com.objective-see.BlockBlockKext

/* FUNCTIONS */

//start
// ->automatically invoked
kern_return_t BlockBlockKext_start(kmod_info_t * ki, void *d);

//stop
// ->automatically invoked
kern_return_t BlockBlockKext_stop(kmod_info_t *ki, void *d);

//automatically invoked when process is exec'd
// ->grab process details and broadcast to user mode
static int processExec(kauth_cred_t cred, struct vnode *vp, struct vnode *scriptvp, struct label *vnodelabel,struct label *scriptlabel, struct label *execlabel,	struct componentname *cnp, u_int *csflags, void *macpolicyattr, size_t macpolicyattrlen);


/* GLOBALS */

//handle to (registered) policy
mac_policy_handle_t policyHandle = {0};

//policy ops
// ->specific MAC hooks
static struct mac_policy_ops bbPolicyOps =
{
    //only interested in exec
    // ->blockblock's policy function
    .mpo_vnode_check_exec = processExec
};

//MAC policy
// ->pretty much empty, save for name and pointer to ops vector
static struct mac_policy_conf policyConf = {
    .mpc_name            = "BB Process Monitor",
    .mpc_fullname        = "BlockBLock Kernel-Mode Process Monitor",
    .mpc_labelnames      = NULL,
    .mpc_labelname_count = 0,
    .mpc_ops             = &bbPolicyOps,
    .mpc_loadtime_flags  = MPC_LOADTIME_FLAG_UNLOADOK, /* TODO: this allows to unload, good idea to REMOVE in release */
    .mpc_field_off       = NULL,
    .mpc_runtime_flags   = 0,
    .mpc_list            = NULL,
    .mpc_data            = NULL
};

//objective-see's vendor id
u_int32_t objSeeVendorID = 0;

//MAC policy function
// ->automatically invoked anytime a new process is started
static int processExec(kauth_cred_t cred, struct vnode *vp, struct vnode *scriptvp, struct label *vnodelabel,struct label *scriptlabel, struct label *execlabel,	struct componentname *cnp, u_int *csflags, void *macpolicyattr, size_t macpolicyattrlen)
{
    //path
    char path[MAXPATHLEN] = {0};
    
    //length
    int pathLength = MAXPATHLEN;
    
    //uid
    uid_t uid = -1;
    
    //pid
    pid_t pid = -1;
    
    //ppid
    pid_t ppid = -1;
    
    //kernel event message
    struct kev_msg kEventMsg = {0};
    
    //status var
    errno_t status = KERN_FAILURE;
    
    //dbg msg
    printf("BLOCKBLOCK KEXT: MAC policy hook invoked\n");
    
    //attempt to get path
    status = vn_getpath(vp, path, &pathLength);
    if(KERN_SUCCESS != status)
    {
        //err msg
        printf("BLOCKBLOCK KEXT ERROR: vn_getpath() failed with %d\n", status);
    
        //bail
        goto bail;
    }
    
    //null-terminate
    // ->just to be safe
    path[MAXPATHLEN-1] = 0x0;
    
    //get UID
    uid = kauth_getuid();
    
    //get pid
    pid = proc_selfpid();
    
    //get ppid
    ppid = proc_selfppid();
    
    //dbg msg
    printf("BLOCKBLOCK KEXT: new process: %s %d/%d/%d\n", path, pid, ppid, uid);
    
    //zer0 out kernel message
    bzero(&kEventMsg, sizeof(struct kev_msg));
    
    //set vendor code
    kEventMsg.vendor_code = objSeeVendorID;
    
    //set class
    kEventMsg.kev_class = KEV_ANY_CLASS;
    
    //set subclass
    kEventMsg.kev_subclass = KEV_ANY_SUBCLASS;
    
    //set event code
    kEventMsg.event_code = PROCESS_BEGAN_EVENT;
    
    //add pid
    kEventMsg.dv[0].data_length = sizeof(pid_t);
    kEventMsg.dv[0].data_ptr = &pid;
    
    //add uid
    kEventMsg.dv[1].data_length = sizeof(uid_t);
    kEventMsg.dv[1].data_ptr = &uid;
    
    //add ppid
    kEventMsg.dv[2].data_length = sizeof(pid_t);
    kEventMsg.dv[2].data_ptr = &ppid;
    
    //add ppid
    kEventMsg.dv[3].data_length = (u_int32_t)strlen(path)+1;
    kEventMsg.dv[3].data_ptr = &path;

    //broadcast msg to user-mode
    status = kev_msg_post(&kEventMsg);
    if(KERN_SUCCESS != status)
    {
        //err msg
        printf("BLOCKBLOCK KEXT ERROR: kev_msg_post() failed with %d\n", status);
        
        //bail
        goto bail;
    }
    
    //happy
    status = KERN_SUCCESS;
    
//bail
bail:

    return status;
}

//start function, automatically invoked
// ->get vendor code, register MAC policy to monitor alerts, etc.
kern_return_t BlockBlockKext_start(kmod_info_t * ki, void *d)
{
    //status var
    kern_return_t status = KERN_FAILURE;
    
    //dbg msg
    printf("BLOCKBLOCK KEXT: starting...\n");
    
    //grab vendor id
    status = kev_vendor_code_find(OBJECTIVE_SEE_VENDOR, &objSeeVendorID);
    if(KERN_SUCCESS != status)
    {
        //err msg
        printf("ERROR: BlockBlock kext failed to get a vendor code (%#x)\n", status);
        
        //indicate failure
        status = KERN_FAILURE;
        
        //bail
        goto bail;
    }

    //register MAC policy
    status = mac_policy_register(&policyConf, &policyHandle, d);
    if(KERN_SUCCESS != status)
    {
        //err msg
        printf("ERROR: BlockBlock kext failed to register MAC policy (%#x)\n", status);
        
        //indicate failure
        status = KERN_FAILURE;
        
        //bail
        goto bail;
    }
    
    //dbg msg
    printf("BLOCKBLOCK KEXT: registered MAC policy\n");
    
    //happy
    status = KERN_SUCCESS;
    
//bail
bail:
    
    return status;
}

//stop function, automatically invoked
// ->should never get called in release mode
kern_return_t BlockBlockKext_stop(kmod_info_t *ki, void *d)
{
    //status
    kern_return_t status = KERN_FAILURE;
    
    //dbg msg
    printf("BLOCKBLOCK KEXT: stopping...\n");
    
    //unregister policy
    status = mac_policy_unregister(policyHandle);
    if(KERN_SUCCESS != status)
    {
        //err msg
        printf("BLOCKBLOCK KEXT ERROR: BlockBlock kext failed to unregister MAC policy (%#x)\n", status);
        
        //indicate failure
        status = KERN_FAILURE;
        
        //bail
        goto bail;
    }
    
    //dbg msg
    printf("BLOCKBLOCK KEXT: unregistered MAC policy\n");
    
    //happy
    status = KERN_SUCCESS;
   
//bail
bail:
    
    return status;
}
